# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# This is based on the specifications of:
# X. Li, K. Tang, M. Omidvar, Z. Yang and K. Qin,
# "Benchmark Functions for the CEC’2013 Special Session and Competition on Large Scale Global Optimization",
# Technical Report, Evolutionary Computation and Machine Learning Group, RMIT University, Australia, 2013.
# and the corresponding source code.

from ._funcs import make_function
